// Copyright (c) 1999-2000 David Muse
// See the COPYING file for more information

#include <config.h>

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif
#include <string.h>
#ifdef HAVE_STRINGS_H
	#include <strings.h>
#endif
#include <sys/socket.h>

#include <server.h>

server::server() {
	clientsock=-1;
	serversockin=-1;
	serversockun=-1;
}

server::~server() {
	closeClientSocket();
	closeServerSocket();
}

int	server::openSockets(unsigned short int inetport, 
				char *unixport, mode_t mode, int backlog) {
	if (openInetSocket(inetport,backlog) && 
		openUnixSocket(unixport,mode,backlog)) {
		return 1;
	}
	return 0;
}

int	server::openInetSocket(unsigned short int port, int backlog) {

	int	retval=1;

	bzero((void *)&serversin,sizeof(serversin));
	serversin.sin_family=AF_INET;
	serversin.sin_port=htons(port);
	serversin.sin_addr.s_addr=htonl(INADDR_ANY);

	if ((serversockin=socket(AF_INET,SOCK_STREAM,0))==-1) {

		cerr << "error: inet socket creation failed: ";
		cerr << strerror(errno) << endl;
		retval=0;

	} else {

		setsockopt(serversockin,SOL_SOCKET,SO_REUSEADDR,NULL,NULL);

		if (bind(serversockin,(struct sockaddr *)&serversin,
						sizeof(serversin))==-1) {

			cerr << "error: inet bind failed: ";
			cerr << strerror(errno) << endl;
			retval=0;
	
		} else if (listen(serversockin,backlog)==-1) {

			cerr << "error: inet listen failed: ";
			cerr << strerror(errno) << endl;
			retval=0;
		}
	}

	if (port==0) {
		struct	sockaddr_in	testsin;
		socklen_t	len=sizeof(testsin);
		bzero((void *)&testsin,len);
		getsockname(serversockin,(struct sockaddr *)&testsin,&len);
		serverportin=(unsigned short int)ntohs(testsin.sin_port);
	} else {
		serverportin=port;
	}

	return retval;
}

int	server::openUnixSocket(char *port, mode_t mode, int backlog) {

	int	retval=1;

	if (!port || (port && !port[0])) {
		return retval;
	}

	unlink(port);
	bzero((void *)&serversun,sizeof(serversun));
	serversun.sun_family=AF_UNIX;
	strcpy(serversun.sun_path,port);

	if ((serversockun=socket(AF_UNIX,SOCK_STREAM,0))==-1) {

		cerr << "error: unix socket creation failed: ";
		cerr << strerror(errno) << endl;
		retval=0;

	} else if (bind(serversockun,(struct sockaddr *)&serversun,
						sizeof(serversun))==-1) {

		cerr << "error: unix bind failed: ";
		cerr << strerror(errno) << endl;
		retval=0;
	
	} else if (listen(serversockun,backlog)==-1) {

		cerr << "error: unix listen failed: ";
		cerr << strerror(errno) << endl;
		retval=0;
	}

	if (serversockun) {
		chmod(serversun.sun_path,mode);
	}

	return retval;
}

int	server::acceptConnection(int timeoutsec, int timeoutusec) {

	int	retval=0;


	// if neither socket is open, return failure
	if (serversockun==-1 && serversockin==-1) {
		return retval;
	}


	// set up the timeout
	timeval	*tv;
	if (timeoutsec>-1) {
		tv=new timeval;
		tv->tv_sec=timeoutsec;
		tv->tv_usec=timeoutusec;
	} else {
		tv=NULL;
	}


	// set up list of sockets to select() on
	fd_set	socklist;
	FD_ZERO(&socklist);
	if (serversockin>-1) {
		FD_SET(serversockin,&socklist);
	}
	if (serversockun>-1) {
		FD_SET(serversockun,&socklist);
	}
	int	largest=serversockin;
	if (serversockun>largest) {
		largest=serversockun;
	}


	// wait for one of the sockets to be ready
	int	selectresult;
	for (;;) {

		selectresult=select(largest+1,&socklist,NULL,NULL,tv);

		// handle timeout
		if (selectresult==0) {

			break;

		// handle non-error, on error, select again
		} else if (selectresult>-1) {


			// figure out which socket to accept() on
			if (serversockun>-1 && 
				FD_ISSET(serversockun,&socklist)) {

				// accept the unix socket connection
				socklen_t	clientsunlen=sizeof(clientsun);
				clientsock=accept(serversockun,
						(struct sockaddr *)&clientsun,
						&clientsunlen);
			} else {
	
				// accept the inet socket connection
				socklen_t	clientsinlen=sizeof(clientsin);
				clientsock=accept(serversockin,
						(struct sockaddr *)&clientsin,
						&clientsinlen);
			}
	
			// handle an accept() failure
			if (clientsock>-1) {
				retval=1;
			}

			break;
		}
	}

	// clean up
	if (tv) {
		delete tv;
	}

	return retval;
}

void	server::closeClientSocket() {
	if (clientsock>-1) {
		close(clientsock);
		clientsock=-1;
	}
}

void	server::closeServerSocket() {
	if (serversockin>-1) {
		close(serversockin);
		serversockin=-1;
	}
	if (serversockun>-1) {
		close(serversockun);
		unlink(serversun.sun_path);
		serversockun=-1;
	}
}

int	server::writeToClient(short number) {
	int	temp=htons(number);
	return (int)write(clientsock,(void *)&temp,sizeof(short));
}

int	server::writeToClient(long number) {
	int	temp=htonl(number);
	return (int)write(clientsock,(void *)&temp,sizeof(long));
}

int	server::writeToClient(float number) {
	return (int)write(clientsock,(void *)&number,sizeof(float));
}

int	server::writeToClient(double number) {
	return (int)write(clientsock,(void *)&number,sizeof(double));
}

int	server::writeToClient(char character) {
	return (int)write(clientsock,&character,sizeof(char));
}

int	server::writeToClient(char *string) {
	return (int)write(clientsock,string,strlen(string));
}

int	server::writeToClient(char *string, int size) {
	return (int)write(clientsock,string,size);
}

int	server::writeToClient(void *buffer, int size) {
	return (int)write(clientsock,(void *)buffer,size);
}

int	server::readFromClient(short *buffer) {
	int	retval=(int)read(clientsock,(void *)buffer,sizeof(short));
	*buffer=ntohs(*buffer);
	return retval;
}

int	server::readFromClient(long *buffer) {
	int	retval=(int)read(clientsock,(void *)buffer,sizeof(long));
	*buffer=ntohl(*buffer);
	return retval;
}

int	server::readFromClient(float *buffer) {
	return (int)read(clientsock,(void *)buffer,sizeof(float));
}

int	server::readFromClient(double *buffer) {
	return (int)read(clientsock,(void *)buffer,sizeof(double));
}

int	server::readFromClient(char *buffer, int size) {
	return (int)read(clientsock,(void *)buffer,size);
}

int	server::readFromClient(void *buffer, int size) {
	return (int)read(clientsock,buffer,size);
}

int	server::readFromClient(strstream *buffer, char *terminator) {
	int	filedescriptor=clientsock;
	#include "readfromfileptr.C"
	return totalread;
}

int	server::getServerSocket() {
	return serversockin;
}

int	server::getClientSocket() {
	return clientsock;
}

unsigned short int	server::getInetServerPort() {
	return serverportin;
}

char	*server::getUnixServerPort() {
	return serversun.sun_path;
}
